extends MeshInstance3D

@export var tracker : UltraleapHandTracking

var material : ShaderMaterial
@onready var offsets : PackedVector3Array = PackedVector3Array()
@onready var widths : PackedFloat32Array = PackedFloat32Array()
@onready var lengths : PackedFloat32Array = PackedFloat32Array()

func _ready():
	material = mesh.material
	
func _process(_delta):
	var frame = tracker.last_frame
	set_hand_transform(frame.left_hand, "left")
	material.set_shader_parameter("left_hand_bones_offset", offsets)
	material.set_shader_parameter("left_hand_bones_width", widths)
	material.set_shader_parameter("left_hand_bones_length", lengths)

func set_hand_transform(hand : UltraleapHand, name : String):
	offsets.clear()
	widths.clear()
	lengths.clear()
	set_finger_transform(hand.thumb, name + "_thumb")
	set_finger_transform(hand.index, name + "_index")
	set_finger_transform(hand.middle, name + "_middle")
	set_finger_transform(hand.ring, name + "_ring")
	set_finger_transform(hand.pinky, name + "_pinky")
	
func set_finger_transform(finger : UltraleapDigit, name : String):
	set_bone_transform(finger.metacarpal, name + "_metacarpal")
	set_bone_transform(finger.proximal, name + "_proximal")
	set_bone_transform(finger.intermediate, name + "_intermediate")
	set_bone_transform(finger.distal, name + "_distal")

func set_bone_transform(bone : UltraleapBone, name : String):
	var pos = (bone.next_joint + bone.prev_joint) / 2
	var length = (bone.next_joint - bone.prev_joint).length()
	var transform_m : Transform3D = Transform3D(global_transform)

	pos = transform_m * pos
	
	# Multiply by quaternion, which is this object rotation, to rotate the hands
	var rot = (quaternion * bone.rotation * Quaternion(Vector3.RIGHT, deg_to_rad(90))).get_euler()
	
	material.set_shader_parameter(name, -pos)
	
	offsets.push_back(-pos)
	widths.push_back(bone.width)
	lengths.push_back(length)
