#include "ultraleap.h"

#include <godot_cpp/core/class_db.hpp>

#include <godot_cpp/classes/global_constants.hpp>
#include <godot_cpp/classes/time.hpp>
#include <godot_cpp/variant/utility_functions.hpp>
#include <godot_cpp/variant/vector3.hpp>
#include <godot_cpp/classes/node3d.hpp>

#include <LeapC.h>
#include <time.h>

#include <thread>
#include <functional>

#include "bone.h"
#include "palm.h"
#include "digit.h"
#include "hand.h"
#include "frame.h"

using namespace godot;

UltraleapHandTracking::~UltraleapHandTracking() {
    if (_isRunning) {
        _isRunning = false;
        messageLoop.join();
    }
    LeapCloseConnection(connectionHandle);
    LeapDestroyConnection(connectionHandle);
}

void UltraleapHandTracking::_bind_methods() {
    UtilityFunctions::print("Binding the methods");

    PropertyInfo tracking_mode_info = PropertyInfo(
        Variant::INT,
        "tracking_mode",
        PROPERTY_HINT_ENUM,
        "Desktop, HMD, Screentop"
    );

    UtilityFunctions::print("Declaring the bound methods");
    // TODO: Fix
    // This isn't used yet but it raises a free(): invalid size error so needs to be fixed
    ClassDB::bind_method(D_METHOD("set_interpolate", "interpolate"), &UltraleapHandTracking::set_interpolate, DEFVAL(true));
    ClassDB::bind_method(D_METHOD("get_interpolate"), &UltraleapHandTracking::get_interpolate);
    ClassDB::add_property("UltraleapHandTracking", PropertyInfo(Variant::BOOL, "interpolate"), "set_interpolate", "get_interpolate");

    ClassDB::bind_method(D_METHOD("start"), &UltraleapHandTracking::start);

    ClassDB::bind_method(D_METHOD("set_last_frame", "last_frame"), &UltraleapHandTracking::set_last_frame);
    ClassDB::bind_method(D_METHOD("get_last_frame"), &UltraleapHandTracking::get_last_frame);

    ClassDB::bind_method(D_METHOD("set_tracking_mode", "tracking_mode"), &UltraleapHandTracking::set_tracking_mode, DEFVAL(0));
    ClassDB::bind_method(D_METHOD("get_tracking_mode"), &UltraleapHandTracking::get_tracking_mode);

    ClassDB::bind_method(D_METHOD("get_left_image"), &UltraleapHandTracking::get_left_image);
    ClassDB::bind_method(D_METHOD("get_right_image"), &UltraleapHandTracking::get_right_image);

    UtilityFunctions::print("Declaring the properties");

    // TODO: Fix
    // From here, we have a "munmap_chunk(): invalid pointer" error
    // I can see a pattern of things here... properties not working

    ClassDB::add_property(
        "UltraleapHandTracking",
        PropertyInfo(
            Variant::OBJECT,
            "last_frame",
            PROPERTY_HINT_RESOURCE_TYPE,
            "UltraleapFrame"
        ),
        "set_last_frame",
        "get_last_frame"
    );

    ClassDB::add_property(
        "UltraleapHandTracking",
        tracking_mode_info,
        "set_tracking_mode",
        "get_tracking_mode"
    );

    UtilityFunctions::print("Declaring the signals");

    // SIGNALS

    ClassDB::add_signal(
        "UltraleapHandTracking",
        MethodInfo(
            "tracking_mode_changed",
            tracking_mode_info
        )
    );

    UtilityFunctions::print("Methods bound");
}

void UltraleapHandTracking::start() {
    UtilityFunctions::print("Starting the tracking");
    LEAP_CONNECTION_CONFIG config;

    // Set connection to multi-device aware
    memset(&config, 0, sizeof(config));
    config.size = sizeof(config);
    config.flags = eLeapConnectionConfig_MultiDeviceAware;

    if(connectionHandle || LeapCreateConnection(&config, &connectionHandle) == eLeapRS_Success) {
        eLeapRS result = LeapOpenConnection(connectionHandle);
        if(result == eLeapRS_Success) {
            _isRunning = true;

            LeapSetPolicyFlags(connectionHandle, eLeapPolicyFlag_Images, 0);
            
            // Create the message poll thread
            messageLoop = std::thread(&UltraleapHandTracking::serviceMessageLoop, std::ref(*this));
        }
    }
}

bool UltraleapHandTracking::get_interpolate() {
  return interpolate;
}

void UltraleapHandTracking::set_interpolate(bool value) {
  interpolate = value;
}

void UltraleapHandTracking::handleConnectionEvent(const LEAP_CONNECTION_EVENT *evt) {
    IsConnected = true;
    UtilityFunctions::print("Connected");

    //LeapCreateClockRebaser(&clockSynchronizer);
}

// void UltraleapHandTracking::handleDeviceEvent(const LEAP_DEVICE_EVENT *device_event){
//     LEAP_DEVICE deviceHandle;
//     //Open device using LEAP_DEVICE_REF from event struct.
//     eLeapRS result = LeapOpenDevice(device_event->device, &deviceHandle);
//     if(result != eLeapRS_Success){
//         printf("Could not open device %s.\n", ResultString(result));
//         return;
//     }

//     //Create a struct to hold the device properties, we have to provide a buffer for the serial string
//     LEAP_DEVICE_INFO deviceProperties = { sizeof(deviceProperties) };
//     // Start with a length of 1 (pretending we don't know a priori what the length is).
//     // Currently device serial numbers are all the same length, but that could change in the future
//     deviceProperties.serial_length = 1;
//     deviceProperties.serial = malloc(deviceProperties.serial_length);
//     //This will fail since the serial buffer is only 1 character long
//     // But deviceProperties is updated to contain the required buffer length
//     result = LeapGetDeviceInfo(deviceHandle, &deviceProperties);
//     if(result == eLeapRS_InsufficientBuffer){
//         //try again with correct buffer size
//         deviceProperties.serial = realloc(deviceProperties.serial, deviceProperties.serial_length);
//         result = LeapGetDeviceInfo(deviceHandle, &deviceProperties);
//         if(result != eLeapRS_Success){
//             printf("Failed to get device info %s.\n", ResultString(result));
//             free(deviceProperties.serial);
//             return;
//         }
//     }
//     //setDevice(&deviceProperties);
//     //if(ConnectionCallbacks.on_device_found){
//     //    ConnectionCallbacks.on_device_found(&deviceProperties);
//     //}

//     free(deviceProperties.serial);
//     LeapCloseDevice(deviceHandle);
// }

// void UltraleapHandTracking::handle_tracking_event(const LEAP_TRACKING_EVENT *tracking_event) {
//     // Ideally we want to have a system that uses interpolation as well since we're rebasing the clocks
// }

void UltraleapHandTracking::serviceMessageLoop() {
    LEAP_CONNECTION_MESSAGE msg;
    eLeapRS result;
    int timeout = 100;
    
    eLeapTrackingMode current_tracking_mode;
    UltraleapTypes::TrackingMode tm = tracking_mode;

    while (_isRunning) {
        result = LeapPollConnection(connectionHandle, timeout, &msg);

        // Polling may have taken some time, re-check exit condition
        if (!_isRunning) {
            break;
        }

        switch (msg.type){
            case eLeapEventType_Connection:
                handleConnectionEvent(msg.connection_event);
                break;
            case eLeapEventType_ConnectionLost:
                //handleConnectionLostEvent(msg.connection_lost_event);
                break;
            case eLeapEventType_Device:
                handle_device_event(msg.device_event);
                break;
            case eLeapEventType_DeviceLost:
                //handleDeviceLostEvent(msg.device_event);
                break;
            case eLeapEventType_DeviceFailure:
                //handleDeviceFailureEvent(msg.device_failure_event);
                break;
            case eLeapEventType_Tracking:
                // TODO: SECURE THIS WITH A MUTEX OR SOMETHING
                if(latest_frame == nullptr) {
                    latest_frame = new LEAP_TRACKING_EVENT;
                };
                *latest_frame = *msg.tracking_event;
                fill_frame_data(last_frame, latest_frame);
                break;
            case eLeapEventType_ImageComplete:
                // Ignore
                break;
            case eLeapEventType_ImageRequestError:
                // Ignore
                break;
            case eLeapEventType_LogEvent:
                //handleLogEvent(msg.log_event);
                break;
            case eLeapEventType_Policy:
                break;
            case eLeapEventType_ConfigChange:
                //handleConfigChangeEvent(msg.config_change_event);
                break;
            case eLeapEventType_ConfigResponse:
                //handleConfigResponseEvent(msg.config_response_event);
                break;
            case eLeapEventType_Image:
                //UtilityFunctions::print("Received an image !");
                //fill_image_data(left_image, msg.image_event->image);
                //fill_image_data(right_image, msg.image_event->image + 1);
                fill_images_data(left_image, right_image, msg.image_event->image);
                break;
            case eLeapEventType_PointMappingChange:
                //handlePointMappingChangeEvent(msg.point_mapping_change_event);
                break;
            case eLeapEventType_TrackingMode:
                // TODO: Check if having this and the policy event use the same path to notify a change in tracking mode
                //handleTrackingModeEvent(msg.tracking_mode_event);
                current_tracking_mode = msg.tracking_mode_event->current_tracking_mode;
                if (current_tracking_mode == eLeapTrackingMode_HMD) {
                    tm = UltraleapTypes::TrackingMode::HMD;
                }
                else if (current_tracking_mode == eLeapTrackingMode_ScreenTop) {
                    tm = UltraleapTypes::TrackingMode::Screentop;
                }
                else if (current_tracking_mode == eLeapTrackingMode_Desktop) {
                    tm = UltraleapTypes::TrackingMode::Desktop;
                }
                
                if (tracking_mode != tm) {
                    emit_signal("tracking_mode_changed", tm);
                }
                
                tracking_mode = tm;
                break;
            case eLeapEventType_LogEvents:
                //handleLogEvents(msg.log_events);
                break;
            case eLeapEventType_HeadPose:
                //handleHeadPoseEvent(msg.head_pose_event);
                break;
            case eLeapEventType_IMU:
                //handleImuEvent(msg.imu_event);
                break;
            case eLeapEventType_DeviceStatusChange:
                break;
            case eLeapEventType_DroppedFrame:
                break;
            case eLeapEventType_Eyes:
                break;
            case eLeapEventType_None:
                break;
            default:
                //discard unknown message types
                UtilityFunctions::print("Unhandled message type");
        } //switch on msg.type
    }
}

Vector3 UltraleapHandTracking::ultraleap_vector3_to_godot_vector3(LEAP_VECTOR* vector) {
    return Vector3(
        vector->x / 1000,
        vector->y / 1000,
        vector->z / 1000
    );
}

Quaternion UltraleapHandTracking::ultraleap_quaternion_to_godot_quaternion(LEAP_QUATERNION* quaternion) {
    return Quaternion(
        quaternion->x,
        quaternion->y,
        quaternion->z,
        quaternion->w
    );
}

void UltraleapHandTracking::fill_bone_data(UltraleapBone* ul_bone, LEAP_BONE* bone) {
    ul_bone->prev_joint = ultraleap_vector3_to_godot_vector3(&bone->prev_joint);
    ul_bone->next_joint = ultraleap_vector3_to_godot_vector3(&bone->next_joint);
    ul_bone->width = bone->width / 1000;
    ul_bone->rotation = ultraleap_quaternion_to_godot_quaternion(&bone->rotation);
}

void UltraleapHandTracking::fill_palm_data(UltraleapPalm* ul_palm, LEAP_PALM* palm) {
    ul_palm->position = ultraleap_vector3_to_godot_vector3(&palm->position);
    ul_palm->stabilized_position = ultraleap_vector3_to_godot_vector3(&palm->stabilized_position);
    ul_palm->velocity = ultraleap_vector3_to_godot_vector3(&palm->velocity);
    ul_palm->normal = ultraleap_vector3_to_godot_vector3(&palm->normal);
    ul_palm->direction = ultraleap_vector3_to_godot_vector3(&palm->direction);
    ul_palm->width = palm->width / 1000;
    ul_palm->orientation = ultraleap_quaternion_to_godot_quaternion(&palm->orientation);
}

void UltraleapHandTracking::fill_digit_data(UltraleapDigit* ul_digit, LEAP_DIGIT* digit) {
    ul_digit->is_extended = digit->is_extended == 1;
    
    fill_bone_data(ul_digit->metacarpal, &digit->metacarpal);
    fill_bone_data(ul_digit->proximal, &digit->proximal);
    fill_bone_data(ul_digit->intermediate, &digit->intermediate);
    fill_bone_data(ul_digit->distal, &digit->distal);
}

void UltraleapHandTracking::fill_hand_data(UltraleapHand* ul_hand, LEAP_HAND* hand) {
    ul_hand->type = hand->type == eLeapHandType_Left ? UltraleapTypes::Chirality::Left : UltraleapTypes::Chirality::Right;
    ul_hand->confidence = hand->confidence;
    ul_hand->visible_time = hand->visible_time;
    ul_hand->pinch_distance = hand->pinch_distance;
    ul_hand->pinch_strength = hand->pinch_strength;
    ul_hand->grab_angle = hand->grab_angle;
    ul_hand->grab_strength = hand->grab_strength;

    fill_palm_data(ul_hand->palm, &hand->palm);

    for (size_t i = 0; i < 5; i++)
    {
        fill_digit_data(ul_hand->digits[i], &hand->digits[i]);
    }

    fill_bone_data(ul_hand->arm, &hand->arm);
}

void UltraleapHandTracking::fill_frame_data(UltraleapFrame* ul_frame, LEAP_TRACKING_EVENT* frame) {
    ul_frame->framerate = frame->framerate;
    ul_frame->is_left_hand_visible = false;
    ul_frame->is_right_hand_visible = false;

    for (size_t i = 0; i < frame->nHands; i++)
    {
        if (frame->pHands[i].type == eLeapHandType_Left) {
            fill_hand_data(ul_frame->left_hand, &frame->pHands[i]);
            ul_frame->is_left_hand_visible = true;
        }
        else {
            fill_hand_data(ul_frame->right_hand, &frame->pHands[i]);
            ul_frame->is_right_hand_visible = true;
        }
    }
}

void UltraleapHandTracking::set_tracking_mode(UltraleapTypes::TrackingMode value) {
    if (value != tracking_mode) {
        //tracking_mode = value;
        uint64_t set_flag = 0;
        uint64_t clear_flag = 0;
        if (value == UltraleapTypes::TrackingMode::Desktop) {
            clear_flag = eLeapPolicyFlag_OptimizeHMD | eLeapPolicyFlag_OptimizeScreenTop;
        }
        else if (value == UltraleapTypes::TrackingMode::HMD) {
            set_flag = eLeapPolicyFlag_OptimizeHMD;
            clear_flag = eLeapPolicyFlag_OptimizeScreenTop;
        }
        else if (value == UltraleapTypes::TrackingMode::Screentop) {
            clear_flag = eLeapPolicyFlag_OptimizeHMD;
            set_flag = eLeapPolicyFlag_OptimizeScreenTop;
        }
        else {
            UtilityFunctions::print("Invalid value for tracking mode");
        }
        LeapSetPolicyFlags(connectionHandle, set_flag, clear_flag);
    }
}

// Might be removed, could be replaced by fill_images_data which fills both images in one loop
void UltraleapHandTracking::fill_image_data(UltraleapImage* ul_image, const LEAP_IMAGE* image) {
    uint32_t width = image->properties.width;
    uint32_t height = image->properties.height;

    if (image->data == nullptr) {
        UtilityFunctions::print("Oops null pointer for the image");
        return;
    }

    PackedInt64Array temp = PackedInt64Array();

    uint32_t tot_pixels = height * width;

    for (uint32_t i = 0; i < tot_pixels / 8; i++) {
        uint64_t* data = (uint64_t*)((uint8_t*)image->data + image->offset);
        uint64_t pixel = data[i];
        temp.push_back(pixel);
    }

    ul_image->data = temp.to_byte_array();
}

void UltraleapHandTracking::fill_images_data(UltraleapImage* left, UltraleapImage* right, const LEAP_IMAGE* images) {
    uint32_t width = images->properties.width;
    uint32_t height = images->properties.height;

    if (images->data == nullptr) {
        UtilityFunctions::print("Oops null pointer for the image");
        return;
    }

    PackedInt64Array temp_left = PackedInt64Array();
    PackedInt64Array temp_right = PackedInt64Array();

    uint32_t tot_pixels = height * width;

    for (uint32_t i = 0; i < tot_pixels / 8; i++) {
        uint64_t* data_left = (uint64_t*)((uint8_t*)images->data + images->offset);
        uint64_t* data_right = (uint64_t*)((uint8_t*)(images + 1)->data + (images + 1)->offset);
        uint64_t pixel_left = data_left[i];
        uint64_t pixel_right = data_right[i];
        temp_left.push_back(pixel_left);
        temp_right.push_back(pixel_right);
    }

    left->data = temp_left.to_byte_array();
    right->data = temp_right.to_byte_array();
}

void UltraleapHandTracking::handle_device_event(const LEAP_DEVICE_EVENT* event) {
    // TODO: Add some checks to see if all the operations succeed
    LEAP_DEVICE device;
    LeapOpenDevice(event->device, &device);
    LeapSubscribeEvents(connectionHandle, device);

    LEAP_DEVICE_INFO deviceInfo;
    memset(&deviceInfo, 0, sizeof(deviceInfo));

    // Use stack memory to allocate for the serial number field.
    char serial[1000];
    memset(serial, 0, sizeof(serial));
    deviceInfo.serial = serial;
    deviceInfo.serial_length = sizeof(serial) - 1;
    deviceInfo.size = sizeof(deviceInfo);

    LeapGetDeviceInfo(device, &deviceInfo);

    UtilityFunctions::print("Found device with ID: ", event->device.id, ", type: ", deviceInfo.pid, ", serial number: ", deviceInfo.serial);
}