#ifndef ULTRALEAP_H
#define ULTRALEAP_H

#include <godot_cpp/classes/node3d.hpp>
#include <godot_cpp/classes/global_constants.hpp>
#include <godot_cpp/classes/viewport.hpp>
#include <godot_cpp/variant/vector3.hpp>
#include <godot_cpp/variant/quaternion.hpp>

#include <godot_cpp/core/binder_common.hpp>

#include <LeapC.h>
#include <thread>

#include "bone.h"
#include "palm.h"
#include "digit.h"
#include "hand.h"
#include "types.h"
#include "frame.h"
#include "image.h"

using namespace godot;

class UltraleapHandTracking : public Node3D {
    GDCLASS(UltraleapHandTracking, Node3D);

public:
    ~UltraleapHandTracking();

    void start();

    UltraleapTypes::TrackingMode tracking_mode = UltraleapTypes::TrackingMode::Desktop;

    bool interpolate = true;

    bool get_interpolate();
    void set_interpolate(bool value);

    LEAP_TRACKING_EVENT* latest_frame = nullptr;

    static void fill_bone_data(UltraleapBone* ul_bone, LEAP_BONE* bone);
    static void fill_digit_data(UltraleapDigit* ul_digit, LEAP_DIGIT* digit);
    static void fill_hand_data(UltraleapHand* ul_hand, LEAP_HAND* hand);
    static void fill_frame_data(UltraleapFrame* ul_frame, LEAP_TRACKING_EVENT* frame);
    static void fill_palm_data(UltraleapPalm* ul_palm, LEAP_PALM* palm);
    static void fill_image_data(UltraleapImage* ul_image, const LEAP_IMAGE* image);
    static void fill_images_data(UltraleapImage* left, UltraleapImage* right, const LEAP_IMAGE* images);

    static Vector3 ultraleap_vector3_to_godot_vector3(LEAP_VECTOR* vector);
    static Quaternion ultraleap_quaternion_to_godot_quaternion(LEAP_QUATERNION* quaternion);

    UltraleapFrame* last_frame = memnew(UltraleapFrame);
    Ref<UltraleapFrame> last_frame_ref = Ref<UltraleapFrame>(last_frame);

    UltraleapImage* left_image = memnew(UltraleapImage);
    Ref<UltraleapImage> left_image_ref = Ref<UltraleapImage>(left_image);

    UltraleapImage* right_image = memnew(UltraleapImage);
    Ref<UltraleapImage> right_image_ref = Ref<UltraleapImage>(right_image);

    Variant get_last_frame() { return last_frame_ref; }
    void set_last_frame(Variant value) { return; }

    UltraleapTypes::TrackingMode get_tracking_mode() { return tracking_mode; }
    void set_tracking_mode(UltraleapTypes::TrackingMode value);

    Variant get_left_image() { return left_image_ref; }
    Variant get_right_image() { return right_image_ref; }

protected:
    static void _bind_methods();

private:
    bool _isRunning = false;
    LEAP_CONNECTION connectionHandle = NULL;
    LEAP_CLOCK_REBASER clockSynchronizer;

    bool IsConnected = false;
    std::thread messageLoop;

    void handleConnectionEvent(const LEAP_CONNECTION_EVENT *evt);
    void serviceMessageLoop();

    void handle_device_event(const LEAP_DEVICE_EVENT* event);
};

#endif